/*
 * vpopq_server
 * server for vpopq_client
 *
 * queries local vpopmail API and some other files to check 
 * whether a given recipient is valid
 *
 * this binary should be driven by tcpserver/(x)inetd
 *
 * <rs@bytecamp.net> R. Schulze, bytecamp GmbH
 */

#include <stdio.h>
#include <strings.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <pwd.h>
#include <unistd.h>
#include <fcntl.h>
#include <ctype.h>

#include "config.h"
#include "vpopmail.h"
#include "vauth.h"

#define QMAIL_DEFAULT	".qmail-default"
#define QMAIL_ASSIGN	"/var/qmail/users/assign"
#define VPOPMAIL_BASEDIR "/var/qmail/vpopmail/domains"
#define PATH_BUFLEN	512
#define TOKENS 		":\n"
#define F_BUFLEN	512

int domain_has_catchall(const char *);
char *get_domain_path(const char *);
int domain_has_qmail_user_ext(const char *,const char *);
void lowercase(char *);

static const char __version[]="vpopq_server/0.7/20090319 <rs@bytecamp.net>";

int main(int argc, const char* argv[])
{
char addr[255];
if(fgets(addr, 255, stdin)==NULL)
	{
	puts("?");
	return 1;
	}

/*
 * seek \n, replace by \0
 */

char *nl=index(addr,'\n');
if(nl==NULL)
	{
	puts("?");
	return 1;
	}
*nl=0;

/* 
 * abort if no @ or more than 1 @
 */
char *at=index(addr,'@');
if( (at==NULL) || (index(at+1,'@')!=NULL) )
	{
	puts("?");
	return 1;
	}

/*
 * lowercase...
 */

lowercase(addr);

/*
 * replace @ by 0; addr => user
 */

*at=0;

char *user=addr;
char *domain=at+1;
int answer=-1;

/* 
 * is user known to vpopmail?
 */
if(vauth_getpw(user,domain))
	{
	answer=1;
	}
else
	{
	/* 
	 * maybe there is an alias?
	 */
	char *alias=valias_select(user,domain);
	if(alias!=NULL)
		{
		answer=1;
		}
	else
		{
		/* 
		 * is domain in users/assign?
		 * get path of domain
		 */
		char *domain_path=get_domain_path(domain);
		if(domain_path==NULL)
			{
			answer=0;
			}
		else
			{
			/* 
			 * does domain have a catchall-account (in .qmail-default) ?
			 * if not: look for .qmail-<user> file
			 */
			answer=domain_has_catchall(domain_path);

			if(answer==0)
				{
				/*
				 * is there a .qmail-<user>[-<ext1>[-<ext2>]] file?
				 */
				answer=domain_has_qmail_user_ext(domain_path,user);
				}
			}
		}
	}

fprintf(stderr,"%s@%s: %d\n",user,domain,answer);

switch(answer)
	{
	case 0: puts("0"); break;
	case 1: puts("1"); break;
	default: puts("?"); break;
	}

return 0;
}

/******************************************************************************
 * get_domain_path(domain)
 * returns the path to the working directory of domain
 *****************************************************************************/

char *get_domain_path(const char *domain)
{
FILE *fh;
char *path=NULL;

if((fh=fopen(QMAIL_ASSIGN,"r"))==NULL) 
	{
	return NULL;
	}

/* users/assign looks like
 * +alias.domain.com-:real.domain.com:89:89:/var/vpopmail/domains/real.domain.com:-::
 */
char line[512];
int done=0;

while((!done) && (fgets(line,512,fh)!=NULL) ) 
	{
	char *token=strtok(line,TOKENS);
	if( (token) && (token[0]=='+') && (strstr(token,domain)==token+1) )
		{
		while( (token!=NULL) && (token!=strstr(token,VPOPMAIL_BASEDIR)) )
			{
			token=strtok(NULL,":");
			}
		if(token!=NULL)
			{
			path=calloc(strlen(token)+1,1);
			strcpy(path,token);
			}
		done=1;
		}
	}

fclose(fh);
return path;
}


/******************************************************************************
 * domain_has_catchall(domainpath)
 *
 * opens .qmail-default and checks whether vpopmail is invoked to drop mail
 * to non-existent users
 *****************************************************************************/

int domain_has_catchall(const char *domain_path)
{
char *qmail_default=calloc(strlen(domain_path)+1+strlen(QMAIL_DEFAULT)+1,sizeof(char));
sprintf(qmail_default,"%s/%s",domain_path,QMAIL_DEFAULT);

struct stat st;
if(stat(qmail_default,&st)==-1)
	{
	free(qmail_default);
	return -1;
	}

int fd=open(qmail_default,O_RDONLY | O_NOFOLLOW);
if(fd==-1)
	{
	free(qmail_default);
	return -1;
	}

char *fbuf=calloc(F_BUFLEN,sizeof(char));
if(!fbuf)
	{
	free(qmail_default);
	return -1;
	}

ssize_t bytesread=read(fd,fbuf,F_BUFLEN-1);
if(bytesread==0)
	{
	free(qmail_default);
	free(fbuf);
	return -1;
	}

close(fd);

*(fbuf+bytesread)=0;

/* fprintf(stderr,"file %s: %s",qmail_default,fbuf); */

int ret=1;
if( (strstr(fbuf,"vdelivermail '' bounce-no-mailbox")!=NULL) || (strstr(fbuf,"vdelivermail '' delete")!=NULL) )
	{
	ret=0;
	}
	
free(qmail_default);
free(fbuf);

return ret;
}

/******************************************************************************
 * domain_has_qmail_user_ext(user, domain)
 * user "foo-bar-baz-bam", check for:
 *	.qmail-foo-bar-baz-bam
 *	.qmail-foo-bar-baz
 *	.qmail-foo-bar
 * 	.qmail-foo
 *****************************************************************************/

int domain_has_qmail_user_ext(const char *domain_path, const char *user)
{
/*
 * username: convert . to :
 */

unsigned int user_len=strlen(user);
char *escaped_user=calloc(user_len+1,1);
strcpy(escaped_user,user);

char *ptr=escaped_user;
while(*ptr!=0)
	{
	if(*ptr=='.')
		{
		*ptr=':';
		}
	ptr++;
	}

ptr=escaped_user;
int finished=0;
int ret=0;

while(ret==0 && finished==0)
	{
	/*
	 * build a string for .qmail-localpart and stat that file
	 * domain_path + '/' + ".qmail-" + user + '\0'
	 */
	unsigned int part_len=strlen(escaped_user);
	char *qmail_user=calloc(strlen(domain_path)+1+7+part_len+1,sizeof(char));
	sprintf(qmail_user,"%s/.qmail-%s",domain_path,escaped_user);

	struct stat st;
	if(stat(qmail_user,&st)==0)
		{
		ret=1;
		}
	else
		{
		/*
	         * find last '-', mark as end of string ("foo-bar-baz\0" => "foo-bar\0")
	  	 */
		ptr=rindex(escaped_user,'-');
		if(ptr==NULL)
			{
			finished=1;
			}
		else
			{
			*ptr=0;
			}
		}
	free(qmail_user);
	}

free(escaped_user);

return ret;
}

/******************************************************************************
 * lowercase(string)
 * converts all chars to lowercase
 *****************************************************************************/

void lowercase(char *str)
{
char *ptr=str;
while(*ptr!=0)
	{
	*ptr=tolower(*ptr);
	ptr++;
	}
}


